package rabinizer.exec;

import java.io.*;
import java.util.*;

import rabinizer.formulas.*;
import rabinizer.parser.LTLParser;

/**
 * Entry point.
 * 
 * @author Andreas Gaiser & Ruslan Ledesma-Garza
 *
 * @param <P>
 */
public class Rabinizer {

	public static void errorMessageAndExit(String text) {
        System.out.println("ERROR: " + text);
        System.exit(1);
	}
	
	public static void printUsage() {
		System.out.println(
"\n"+
"Usage: rabinizer [options] <ltlfile>\n"+
"   Read LTL[F,G,X,U]-formula from file <ltlfile> and\n"+
"   compute the corresponding Gen Rabin automaton\n" +
"   and the corresponding Rabin automaton.\n"+
"   By default, dump each constructed item to a corresponding\n"+
"   file <ltlfile.*>.\n" +
"Options:\n"+
"   -h          : print this help and exit\n"+
"   -v          : verbose terminal output\n"+
"   -nodump     : only construct, do not dump results to files\n"+
"   -nodot      : (deprecated) same as -nodump\n"+
"   -ltl2dstar  : dump formula in reverse polish notation and exit\n");
	}

    /**
     * @param args
     */
    public static void main(String[] args) {
        
        System.out.println(
"******************************************************************************\n" +
"* Rabinizer 2                                                                *\n" +
"******************************************************************************\n" +
"* The constructor of deterministic Rabin automata for LTL-formulae           *\n" +
"* Release: Mon Feb 3 11:50:56 CEST 2013                                      *\n" +
"* Version 1 by Andreas Gaiser                                                *\n" +
"* Version 2 by Ruslan Ledesma-Garza                                          *\n" +
"* Version 3 by Jan Kretinsky                                                 *\n" +
"******************************************************************************");
                
        // Phase 0: Parsing of arguments
        String ltlfile = null;
        boolean dump = true;
        boolean ltl2dstar = false;
        System.out.print("Arguments:");
        for (String arg : args) {
            System.out.print(" "+arg);
            if (arg.equals("-h") ||
            	arg.equals("--h") || 
                arg.equals("-help") ||
                arg.equals("--help") ) {
                                                       printUsage();
                                                       System.exit(0);
            } else if (arg.equals("-v") ||
            	arg.equals("--v") || 
                arg.equals("-verbose") ||
                arg.equals("--verbose") )              Misc.verbose = true;
            else if (arg.equals("-nodump") ||
                     arg.equals("--nodump") ||
                     arg.equals("-nodot") ||
                     arg.equals("--nodot"))         dump = false;
            else if (arg.equals("-ltl2dstar") ||
                     arg.equals("--ltl2dstar"))     ltl2dstar = true;
            else if (arg.substring(0, 1).equals("-")) {
                                                       System.out.println("\n\nERROR: unknown option " + arg);
                                                       printUsage();
                                                       System.exit(1);
            } else                                     ltlfile = arg;
        }
        System.out.println("");
        if (ltlfile == null) errorMessageAndExit("No input file was given.");
                    
        // Phase 1: Parsing of the input formula
        Misc.verboseln("Phase 1: Parsing of the input formula");
        long start_time_parsing = System.currentTimeMillis();
        LTLParser parser = null;
        try { parser = new LTLParser(new FileReader(new File(ltlfile))); }
        catch (Exception e) {
            errorMessageAndExit(
                          "Exception when opening the input file: "
                          + e.getLocalizedMessage()); }
        Formula phi = null;
        try { phi = parser.parse(); }
        catch (Exception e) {
            errorMessageAndExit("Exception when parsing: "
                                + e.getLocalizedMessage()); }
        System.out.println("Input formula: " + phi.toString());
        long interval_parsing = System.currentTimeMillis() - start_time_parsing;

        // Phase 1a: Dump ltl2dstar formula
        if (ltl2dstar) {
        	System.out.println("ltl2dstar format: "+phi.toReversePolishString());
        	System.exit(0);
        }
        
        // Phase 1b: Dump NNF formula
        phi = phi.toNNF();
        System.out.println("Input formula in NNF: "+phi);
        
        // Phase 1c: Check that the formula is in the fragment
        /*if (phi.untilInsideForall())
        	errorMessageAndExit("The formula is not in the fragment: there is an occurrence of until operator\n" +
        			"in a parameter of always operator.");
        */
        
        // Phase 2: Enumeration of valuations
        Misc.verboseln("Phase 2: Enumeration of valuations");
        long start_time_valuations = System.currentTimeMillis();
        Misc.initializeValuations(Misc.bijectionIdAtom.size());
        long interval_valuations = System.currentTimeMillis() - start_time_valuations;
        
        // Phases 3 & 4: segment automata & dfa
        DFA dfaA = null;
        long start_time_segment_automata, interval_segment_automata;
        long start_time_dfa, interval_dfa;
        
        /*
        // Go either for the old method 
        // - with one step history and
        // - with one state per valuation
        switch(method) {
       	case 0:
        	// Phase 3a: Construction of history automaton
       		start_time_segment_automata = System.currentTimeMillis();
	        List<SegmentAutomaton> historyAutomaton = new ArrayList<SegmentAutomaton>(1);
	        SegmentAutomaton hSA = new SegmentAutomaton(Misc.valuations);
	        historyAutomaton.add(hSA);
	        // Phase 3b: Construction of initial segment state
	        List<Set<Formula>> initialSegmentState = new ArrayList<Set<Formula>>(1);
	        Set<Formula> initSS = new HashSet<Formula>();
	        initSS.add(hSA.initialState);
	        initialSegmentState.add(initSS);
	        interval_segment_automata = System.currentTimeMillis() - start_time_segment_automata;	        
	        	        
	        // Phase 4: Construction of DFA A
	        start_time_dfa = System.currentTimeMillis();
	        dfaA = new DFA(phi, historyAutomaton, initialSegmentState, Misc.valuations);
	        interval_dfa = System.currentTimeMillis() - start_time_dfa;

	    break;
        // ...or for the old method 
        // - with one step history and
        // - with collapsing of states...
       	case 1:
	        throw new Error("Rabinizer.main(): method 1 NYI");
        // break;

        // ...or the new method with segment automata
       	default:
       	*/
        	
       		// Phase 3: Construction of segment automata
        	Misc.verboseln("Phase 3: Construction of segment automata");
       		
        	// Phase 3a: Construction of always- and eventually- recurrent formulas aRF1, ..., aRFm, eRF1, ..., eRFn
        	start_time_segment_automata = System.currentTimeMillis();
        	//Tuple<Set<Formula>, Set<Formula>> recurrentFormulas = phi.recurrentFormulas();
	        List<Formula> alwaysRecFs = new ArrayList<Formula>();
	        alwaysRecFs.addAll(phi.argumentsAlways()); //(recurrentFormulas.left);
	        //List<Formula> eventuallyRecFs = new ArrayList<Formula>();
	        //eventuallyRecFs.addAll(recurrentFormulas.right);
	        // Phase 3b: Construction of always segment automata aSA1, ..., aSAm and their corresponding initial segment states
	        List<SegmentAutomaton> alwaysSAs = new ArrayList<SegmentAutomaton>(alwaysRecFs.size());
	        List<Set<Formula>> alwaysISSs = new ArrayList<Set<Formula>>(alwaysRecFs.size());
	        for (Formula aRF : alwaysRecFs) {
	        	// segment automaton
	        	Misc.verboseln("\nAlways segment automaton");
	        	SegmentAutomaton aSA = new SegmentAutomaton(aRF, Misc.valuations);
	        	alwaysSAs.add(aSA);
	        	// corresponding initial segment state
	        	Set<Formula> iSS = new HashSet<Formula>();
	        	if (aSA.finalStates.contains(Misc.fTrue)) {
	        		iSS.add(Misc.fTrue);
	        	} else {
	        		boolean pickedFS = false;
	        		for(Formula f : aSA.finalStates) {
	        			if (!f.equals(Misc.fFalse)) {
	        				iSS.add(f);
	        				pickedFS = true;
	        				break;
	        			}
	        		}
	        		if (!pickedFS) iSS.add(Misc.fFalse);
	        	}
	        	Misc.verboseln("(always) Final state for iSS: "+iSS);
	        	iSS = aSA.backwardsReachability(iSS);
	        	Misc.verboseln("(always) Initial segment state: "+iSS);
	        	alwaysISSs.add(iSS);
	        }
	        // Phase 3c: Construction of eventually segment automata eSA1, ..., eSAm and their corresponding initial segment states
	        /*
	        List<SegmentAutomaton> eventuallySAs = new ArrayList<SegmentAutomaton>(eventuallyRecFs.size());
	        List<Set<Formula>> eventuallyISSs = new ArrayList<Set<Formula>>(eventuallyRecFs.size());
	        for (Formula eRF : eventuallyRecFs) {
	        	// segment automaton
	        	Misc.verboseln("\nEventually segment automaton");
	        	SegmentAutomaton eSA = new SegmentAutomaton(eRF, Misc.valuations);
	        	eventuallySAs.add(eSA);
	        	/// corresponding initial segment state
	        	Set<Formula> iSS = new HashSet<Formula>();
	        	if (eSA.finalStates.contains(Misc.fFalse)) {
	        		iSS.add(Misc.fFalse);
	        	} else {
	        		boolean pickedFS = false;
	        		for(Formula f : eSA.finalStates) {
	        			if (!f.equals(Misc.fTrue)) {
	        				iSS.add(f);
	        				pickedFS = true;
	        				break;
	        			}
	        		}
	        		if (!pickedFS) iSS.add(Misc.fTrue);
	        	}
	        	Misc.verboseln("(always) Final state for iSS: "+iSS);
	        	iSS = eSA.backwardsReachability(iSS);
	        	eventuallyISSs.add(iSS);
	        	Misc.verboseln("(always) Initial segment state: "+iSS);
        	}
            */     
	        interval_segment_automata = System.currentTimeMillis() - start_time_segment_automata;

	        // Phase 4: Construction of DFA A
	        Misc.verboseln("Phase 4: Construction of Generalized Rabin Automaton");
	        start_time_dfa = System.currentTimeMillis();
	        //alwaysSAs.addAll(eventuallySAs);
	        List<SegmentAutomaton> segmentAutomata = alwaysSAs;
	        //alwaysISSs.addAll(eventuallyISSs);
	        List<Set<Formula>> initialSegmentStates = alwaysISSs;
	        Misc.verboseln("");
	        dfaA = new DFA(phi, segmentAutomata, initialSegmentStates, Misc.valuations);
	        interval_dfa = System.currentTimeMillis() - start_time_dfa;
	        System.out.println("\nGeneralized Rabin Automaton: ");
	        System.out.println("\tNumber of states    : " + dfaA.states.size());
	        System.out.println("\tNumber of edges     : " + dfaA.edges.size());
/*	        
        }
*/
        /*
        // Phase 5: Construction of Generalized Rabin Acceptance Condition
        Misc.verboseln("Phase 5: Construction of Generalized Rabin Acceptance Condition");
        long start_time_grac = System.currentTimeMillis();
    	Misc.verboseln("");
    	List<Formula> subformulasAlways = new ArrayList<Formula>();
		subformulasAlways.addAll(phi.argumentsAlways());
		List<Formula> subformulasExist = new ArrayList<Formula>();
		subformulasExist.addAll(phi.argumentsFinsideG(false));
		Misc.verboseln("subformulasExist: "+subformulasExist);
		GeneralizedRabinAcceptanceCondition genRabinAC = new GeneralizedRabinAcceptanceCondition(subformulasAlways, subformulasExist, dfaA);
    	long interval_grac = System.currentTimeMillis() - start_time_grac;
        System.out.println("\tSize of Gen Rabin AC: " + genRabinAC.size());
    	
    	// Phase 6: Construction of the Rabin Automaton
    	Misc.verboseln("Phase 6: Rabin Automaton");
    	long start_time_rabin = System.currentTimeMillis();
    	RabinAutomaton rabin = new RabinAutomaton(dfaA, genRabinAC);
    	long interval_rabin = System.currentTimeMillis() - start_time_rabin;
        System.out.println("\n\nRabin Automaton: ");
        System.out.println("\tNumber of states    : " + rabin.states.size());
        System.out.println("\tNumber of edges     : " + rabin.edges.size());

    	// Phase 7: Construction of Rabin Acceptance Condition
    	Misc.verboseln("");
    	Misc.verboseln("Phase 7: Construction of Rabin Acceptance Condition");
    	long start_time_rabin_ac = System.currentTimeMillis();
    	RabinAcceptanceCondition rabinAC = new RabinAcceptanceCondition(rabin.condition);
    	long interval_rabin_ac = System.currentTimeMillis() - start_time_rabin_ac;
        System.out.println("\tSize of Rabin AC    : " + rabinAC.size());
        */
    		        
        // Phase 7: Dumping of files
    	Misc.verboseln("Phase 7: Dumping files");
    	long start_time_output = System.currentTimeMillis();
        if (dump) {
	        try {
	            FileWriter fw = new FileWriter(new File(ltlfile+".grabin.dot"));
	            System.out.println("\n");
	            System.out.println("Dumping Generalized Rabin Automaton");
	            fw.write(dfaA.toDotty());
	            fw.close();
	        } catch (IOException e) {
	            errorMessageAndExit("IO exception when writing file "+ltlfile+".grabin.dot: " + e.getMessage()); }
	        /*
	        try {
	            FileWriter fw = new FileWriter(new File(ltlfile+".grac"));
	            System.out.println("\n");
	            System.out.println("Dumping Generalized Rabin Acceptance Condition");
	            fw.write(genRabinAC.toString());
	            fw.close();
	        } catch (IOException e) {
	            errorMessageAndExit("IO exception when writing file "+ltlfile+".grac: " + e.getMessage()); }
            try {
	            FileWriter fw = new FileWriter(new File(ltlfile+".rabin.dot"));
	            System.out.println("\n");
	            System.out.println("Dumping Rabin Automaton");
	            fw.write(rabin.toDotty());
	            fw.close();
	        } catch (IOException e) {
	            errorMessageAndExit("IO exception when writing file "+ltlfile+".rabin.dot: " + e.getMessage()); }
            try {
	            FileWriter fw = new FileWriter(new File(ltlfile+".rac"));
	            System.out.println("\n");
	            System.out.println("Dumping Rabin Acceptance Condition");
	            fw.write(rabinAC.toString());
	            fw.close();
	        } catch (IOException e) {
	            errorMessageAndExit("IO exception when writing file "+ltlfile+".rac: " + e.getMessage()); }
            */
            int i = 0;
            Misc.verboseln("\n");
            for (SegmentAutomaton sA : segmentAutomata) {
                try {
                    FileWriter fw = new FileWriter(new File(ltlfile+".segment"+(++i)+".dot"));
                    System.out.println("");
                    System.out.println("Dumping Segment Automaton #"+i);
                    fw.write(sA.toDotty());
                    fw.close();
                } catch (IOException e) {
                    errorMessageAndExit("IO exception when writing file "+ltlfile+".segment"+i+".dot:" + e.getMessage()); }
            }
        }
        long interval_output = System.currentTimeMillis() - start_time_output;
        
        
        // End of the program
        System.out.println("\n\nSUMMARY\n******************************************************************************");
        System.out.println("\nGeneralized Rabin Automaton: ");
        System.out.println("\tNumber of states    : " + dfaA.states.size());
        System.out.println("\tNumber of edges     : " + dfaA.edges.size());
        /*
        System.out.println("\tSize of Gen Rabin AC: " + genRabinAC.size());
        System.out.println("\n\nRabin Automaton: ");
        System.out.println("\tNumber of states    : " + rabin.states.size());
        System.out.println("\tNumber of edges     : " + rabin.edges.size());
        System.out.println("\tSize of Rabin AC    : " + rabinAC.size());
        */
        System.out.println(
                           "\n\nTotal time                                 : "
                           + (interval_parsing
                        		   + interval_valuations 
                        		   + interval_segment_automata
                        		   + interval_dfa
          //              		   + interval_grac
          //              		   + interval_rabin
          //              		   + interval_rabin_ac
                        		   + interval_output)
                           + "msec");
        if (dump) 
        System.out.println(
                           "Total time minus output to files           : "
                           + (interval_parsing
                           + interval_valuations 
             		       + interval_segment_automata
             		       + interval_dfa
        //     		       + interval_grac
        //     		       + interval_rabin
        //        		   + interval_rabin_ac
             		       )
                + "msec");
        System.out.println(
                           "Parsing of input formula                   : "
                           + interval_parsing + "msec");
        System.out.println(
                	       "Enumeration of valuations                  : "
                		   + interval_valuations + "msec");
        System.out.println(
                           "Construction of segment automata           : "
                           + interval_segment_automata + "msec");
        System.out.println(
        		           "Construction of DFA                        : "
        		           + interval_dfa + "msec");
        /*
        System.out.println(
		                   "Construction of Gen Rabin Accept Condition : "
		                   + interval_grac + "msec");
        System.out.println(
                           "Construction of Rabin Automaton            : "
                           + interval_rabin + "msec");
        System.out.println(
                           "Construction of Rabin Acceptance Condition : "
                           + interval_rabin_ac + "msec");
        */
        if (dump) System.out.println(
		                   "Output to files                            : "
		                   + interval_output + "msec");

    }
}
